#include "StdAfx.h"
#include "langobjs.h"
#include "VariableManager.h"
#include "parser.h"
#include "IfStatementBlock.h"
#include "FunctionCall.h"
#include "asmcommand.h"
#include "labelcommand.h"
#include "gotocommand.h"

CVariableManager *mainmanager;
ostream * out;

//Changes

/*Kyle Stemen - 1/9/98
  I changed all occurances of Program *main; to Program *prog; along with the
  accesses to the variable.
  Program::functions and Program::functionnum made protected.
  CString To68kAsm added to all classes which it applies to.
  added couts in main for the asm to be outputed. It should later be saved into
	a file.
  Fixed how GetVariableType didn't return anything if it wasn't any of the
	variable types.
  Added varVoid, and code to support it.
*/

/*Kyle Stemen - 6/11/98
I am working on converting the coding style to Visual C++.
This will eventually involve using CArray templates, and
CList templates.
*/

/*Kyle Stemen - 6/13/98
I had to give up using a true console application because I would not
then be able to use CString. So instead I used some of the classes
I had already made to write my own console. All of the files that
have nothing to do with with the console application and are mostly
standard between compilers are in the standard directory.
*/

//Explanations and Plans

//Register Optiminization
/*For now my current plan is to keep the compiler using 2 cycles. The change
  from the java version of my code is that the Exection or to68kasm cycle will
  start with the main function, instead of which ever function was typed first.
  From there everytime that function or other subfunctions need to call another
  function it will initiate the preto68kasm cycle in that sub-function. The
  register that would be most convienent to the calling function will be known
  by the VariableManager. The sub-function will return which registers it ended
  up using. The preto68kasm function will save its asm code so that it does not
  recalculate the assembly code, and so the same code can be returned for
  to68kasm.*/

CVariable *GetVariable(CBlock *curblock,CProgram *prog,CString var,
	BOOL inasm)
{
	if(var=="")
		return 0;
	//if(saveto&&((!oldersibling)||oldersibling->saveto!=saveto))
	//	return saveto;
	/*if(child)
	{
		child->saveto=usaveto;
		saveto=child->DecodeExpression(curblock,prog);
		return saveto;
	}*/
	CVariable *ret=curblock->FindVariable(var);
	if(ret)
		return ret;
	return new CVariable(curblock,var,inasm);
}

CString A68kOpcode(CString opcode,CVariable *source,CVariable *dest,
				   CString comments)
{
	comments.TrimLeft();
	if(comments!=""&&comments[0]!=';')
		comments="; "+comments;
	CString asmcode="";
	if(source==0&&dest==0)
		return opcode;
	if(opcode=="addi")
		opcode="add";
	if(source->constant)
	{
		if(dest)
			asmcode=dest->Load();
		if(dest->LoadedInRegister())
		{
			if(opcode=="add")
			{
				if(comments=="")
					comments="; "+dest->GetCName()+
					" += "+source->GetCName()+";";
				asmcode+=FormatCode("addi"+
					dest->GetLengthIdentifier(),
					source->GetName()+","+
					dest->GetRegisterAccess(),comments);
				return asmcode;
			}
			if(opcode=="move")
			{
				if(comments=="")
					comments="; "+dest->GetCName()+
					" = "+source->GetCName()+";";
				asmcode+=FormatCode("moveq",
					source->GetName()+","+
					dest->GetRegisterAccess(),comments);
				return asmcode;
			}
		}
	}
	CRegisterHolder regs;
	regs=vmanager->regs;
	CString temp;
	if(dest)
		temp=dest->Load(1);
	//asmcode+=vmanager->BackupVariables(parent,regs);
	asmcode+=temp;

	regs=vmanager->regs;
	if(source)
		if(opcode=="move"&&dest->GetAccess()!="")
		{
			temp=source->Load(0,dest->GetAccess()[1]-48);
		}
		else
			temp=source->Load();
	else
		temp="";
	//asmcode+=vmanager->BackupVariables(parent,regs);
	asmcode+=temp;
	CString access;
	if(source->constant)
		access=source->GetName();
	else
		access=source->GetAccess();
	if(opcode=="add")
	{
		if(comments=="")
			comments="; "+dest->GetCName()+
			" += "+source->GetCName()+";";
		asmcode+=FormatCode("add"+
			dest->GetLengthIdentifier(),
			access+","+
			dest->GetAccess(),comments);
		return asmcode;
	}
	if(opcode=="move")
	{
		if(dest->GetAccess()==source->GetAccess())
			//because of newline
			return asmcode.Mid(0,asmcode.GetLength()-1);
		if(comments=="")
			comments="; "+dest->GetCName()+
			" = "+source->GetCName()+";";
		if(source->GetName()=="&")
		{
			asmcode+=FormatCode("lea",
				source->initialvalue->GetAccess()+","+
				dest->GetAccess(),comments);
		}
		else
		{
			asmcode+=FormatCode("move"+
				dest->GetLengthIdentifier(),
				access+","+
				dest->GetAccess(),comments);
		}
		return asmcode;
	}
	else
	{
		if(comments=="")
			comments="; Warning: unknown opcode, but I will try.";
		if(dest)
		{
			if(opcode.Find(".")==-1)
				opcode+=dest->GetLengthIdentifier();
			asmcode+=FormatCode(opcode,
				access+","+
				dest->GetAccess(),comments);
		}
		else
		{
			asmcode+=FormatCode(opcode,access,comments);
		}
		return asmcode;
	}
	return asmcode;
}

CParser::CParser(ostream *pOut,CString fname)
{
	out=pOut;
	SetFile(fname);
}

void CParser::StartParsing()
{
	curfunc=0;
	prog=new CProgram();
	curblock=prog;
	CString todecode=" ";
	while(todecode!="")
	{
		todecode=scanner->GetNonspaceToken();
		gout << todecode;
		if(CheckPreprocessorDirective(todecode)) continue;
		int vartype=GetVariableType(todecode);
		if(vartype!=varNone)
		{
			todecode=scanner->GetNonspaceToken();
			gout << " " << todecode;
			CString str2=scanner->GetNonspaceToken();
			gout << str2;
			if(CheckFunction(vartype,todecode,str2)) continue;
			if(CheckVariable(vartype,todecode,str2)) continue;
		}
	}
}

void CParser::ParseBlock(CBlock* ncurblock)
{
	CBlock *ocurblock=curblock;
	curblock=ncurblock;
	CString todecode=" ";
	CString str2;
	while(todecode!="}"&&todecode!="")
	{
		todecode=scanner->GetNonspaceToken();
		gout << todecode;
		if(todecode=="}") continue;
		if(CheckPreprocessorDirective(todecode)) continue;
		if(CheckIfStatement(todecode)) continue;
		if(CheckAsmcode(todecode)) continue;
		CVariable *var;
		if(todecode=="*")
		{
			todecode=scanner->GetNonspaceToken();
			var=GetVariable(todecode);
			var=new CVariable(curblock,var->type,"*",0,var);
		}
		else
			var=GetVariable(todecode);
		str2="";
		if(var)
		{
			str2=scanner->GetNonspaceToken();
			if(CheckAssignment(var,str2)) continue;
			if(CheckAddto(var,str2)) continue;
		}
		int vartype=GetVariableType(todecode);
		if(vartype!=varNone)
		{
			CString todecode2;
			if(str2=="")
			{
				todecode2=scanner->GetNonspaceToken();
				gout << " " << todecode2;
			}
			else
			{
				todecode2=str2;
			}
			str2=scanner->GetNonspaceToken();
			gout << str2;
			if(CheckVariable(vartype,todecode2,str2)) continue;
		}
		if(str2="")
			str2=scanner->GetNonspaceToken();
		if(CheckFunctionCall(todecode,str2)) continue;
		if(CheckLabel(todecode,str2)) continue;
		if(CheckGoto(todecode,str2)) continue;
		gout << "Error: unknown token '" << todecode << "'!\n";
	}
   curblock=ocurblock;
}

char CParser::CheckPreprocessorDirective(CString str)
{
	if(str=="#")
	{
		gout << "Its a preprocessor directive" << endl;
		CString todecode=scanner->GetToken();
		gout << todecode << endl;
		if(CheckInclude(todecode)) return 1;
		return 1;
	}
	return 0;
}

char CParser::CheckInclude(CString str)
{
	if(str!="include") return 0;
	gout << "Start include statement." << endl;
	CString todecode=scanner->GetNonspaceToken();
	gout << todecode;
	if(todecode!="<"&&todecode!="\"")
	{
		gout << "Error! A < or \" is expected!" << endl;
		return 1;
	}
	todecode=scanner->GetNonspaceToken();
	gout << todecode;
	todecode=scanner->GetNonspaceToken();
	gout << todecode;
	if(todecode==".")
	{
		todecode=scanner->GetNonspaceToken();
		gout << todecode;
		todecode=scanner->GetNonspaceToken();
		gout << todecode;
	}
	if(todecode!=">"&&todecode!="\"")
	{
		gout << "Error! A > or \" is expected! You put in a " << todecode << endl;
		return 1;
	}
	gout << endl << "End include statement." << endl;
	return 1;
}

char CParser::CheckIfStatement(CString str)
{
	if(str!="if") return 0;
	gout << "Its an if statement!" << endl;
	CString todecode=scanner->GetNonspaceToken();
	if(todecode!="(")
	{
		gout << "!Error: ( expected!" << endl;
		return 1;
	}
	CExpressionObject exp(scanner,")",(CExpressionObject *)0,"");
	CIfStatementBlock *ifblock=new CIfStatementBlock(curblock,prog,exp.DecodeForIf(curblock,prog));
	curblock->AddItem(ifblock);

	todecode=scanner->GetNonspaceToken();
	if(todecode=="{")
	{
		gout << "If started." << endl;
		ParseBlock(ifblock);
		gout << "If ended." << endl;
	}
	else
	{
   		gout << "Error this part is not finished. For now a { should be here instead of " << todecode << "." << endl;
	}
	return 1;
}

char CParser::CheckFunction(int vartype,CString str1,CString str2)
{
	if(vartype==varNone) return 0;
	if(!scanner->CheckLetterOrNumber(str1[0])) return 0;
	if(str2!="(")
		return 0;
	gout << endl << "Determined to be a function." << endl;
	if(str1=="main")
		curfunc=new CFunction(prog,"_main");
	else
		curfunc=new CFunction(prog,str1);
	CString todecode=" ";
	CString str3,str4;
	while(todecode!=")")
	{
		todecode=scanner->GetNonspaceToken();
		gout << todecode;
		if(todecode==")")
			break;
		int vartype=GetVariableType(todecode);
		if(vartype==varNone)
		{
			gout << "Error: variable type expected instead of '" << todecode <<
				"'." << endl;
			return 0;
		}
		str3=scanner->GetNonspaceToken();
		str4=scanner->GetNonspaceToken();
		if(str3[0]=='*')
		{
			str3=str4;
			str4=scanner->GetNonspaceToken();
			if(str4=="=")
			{
				gout << "Error: not yet implemented!." << endl;
			}
			else if(str4==","||str4==")")
			{
				gout << "Determined to be a pointer parameter." << endl;
				CVariable *var=new CVariable(curfunc,vartype,str3,-1,0);
				curfunc->AddParameter(var);
			}
		}
		else if(!scanner->CheckLetterOrNumber(str3[0]))
		{
			gout << "Error: a letter is expected here." << endl;
			return 0;
		}
		else if(str4=="=")
		{
			gout << "Error: not yet implemented." << endl;
		}
		else if(str4==","||str4==")")
		{
			gout << "Determined to be a parameter." << endl;
			CVariable *var=new CVariable(curfunc,vartype,str3,0,0);
			curfunc->AddParameter(var);
		}
		todecode=str4;
	}
	prog->AddItem(curfunc);
	todecode=scanner->GetNonspaceToken();
	if(todecode!="{")
	{
		gout << "Error! A { expected!" << endl;
		return 1;
	}
	gout << endl;
	ParseBlock(curfunc);
	gout << endl << "End Function" << endl;
	return 1;
}

char CParser::CheckVariable(int vartype,CString str1,CString str2)
{
	if(vartype==varNone)
	{
		gout << "!Error" << endl;
		return 0;
	}
	if(str1[0]=='*')
	{
		str1=str2;
		str2=scanner->GetNonspaceToken();
		if(str2=="=")
		{
			gout << "Error: not yet implemented!." << endl;
		}
		else if(str2==";")
		{
			gout << "Determined to be a pointer." << endl;
			CVariable *var=new CVariable(curblock,vartype,str1,-1,0);
			curblock->AddItem(var);
		}
		return 1;
	}
	else if(!scanner->CheckLetterOrNumber(str1[0]))
	{
		gout << "!Error" << endl;
		return 0;
	}
	if(str2=="=")
	{
		gout << "Determined to be a variable with assignment." << endl;
		//string assignment=scanner->GetNonspaceToken();

		//curblock->AddItem(new CVariable(curblock,vartype,str1,0,
		//	new CVariable(curblock,assignment)));
		CVariable *var=new CVariable(curblock,vartype,str1,0,0);
		curblock->AddItem(var);
		DecodeExpression(";",var,"=");

		/*string semicolon=scanner->GetNonspaceToken();
		if(semicolon!=";")
		{
			cout << "!Error ';' expected instead of " << semicolon << " ." << endl;
		}*/
		return 1;
	}
	else if(str2==";")
	{
		gout << "Determined to be a variable." << endl;
		CVariable *var=new CVariable(curblock,vartype,str1,0,0);
		curblock->AddItem(var);
		return 1;
	}
	return 0;
}

char CParser::CheckAssignment(CVariable *var,CString str1)
{
	if(!var)
		return 0;
	if(str1!="=")
		return 0;
	gout << "Its a variable Assignment!" << endl;
	DecodeExpression(";",var,"=");
	//curblock->AddItem(new CAssignCommand(curblock,prog,
	//	var,str1));
	return 1;
}

char CParser::CheckAddto(CVariable *var,CString str1)
{
	if(!var)
		return 0;
	if(str1!="+=")
		return 0;
	gout << "Its a variable Addto!" << endl;
	CVariable *dest=DecodeExpression(";",var,"+");
	//curblock->AddItem(new AddtoCommand(curblock,prog,
	//	var,dest));
	return 1;
}

int CParser::GetBiggerType(CVariable *var1,CVariable *var2)
{
	if(var1->GetType()>var2->GetType())
		return var1->GetType();
	return var2->GetType();
}

int CParser::GetVariableType(CString str)
{
	if(str=="void") return varVoid;
	if(str=="char") return varChar;
	if(str=="int") return varInt;
	if(str=="long") return varLong;
	return varNone;
}

CVariable *CParser::GetVariable(CString str)
{
	return curblock->FindVariable(str);
}

CVariable *CParser::DecodeExpression(CString end,CVariable *op1,CString newop,CString end2)
{
	if(newop!="=")
	{
		gout << "newop=" << newop;
	}
	if(op1&&newop!="")
	{
		CExpressionObject exp(scanner,end,op1,newop,0,end2);
		return exp.DecodeExpression(curblock,prog);
	}
	CExpressionObject exp(scanner,end,(CExpressionObject *)0,end2);
	return exp.DecodeExpression(curblock,prog);
}

/*Old Decode Expression
Variable *Parser::DecodeExpression(string end,Variable *op1,string newop)
{
	Variable *operand1=op1;
	string **operands;
	CString **operators;
	CString str=scanner->GetNonspaceToken();
	int operationnum=0;
	while(str!=end)
	{
		if(operationnum%5==0)
		{
			operands=new CString*[operationnum+5];
			operators=new CString*[operationnum+5];
		}
		operands[operationnum]=
	Variable *operand2=0;
	CString operation=newop;
	char firstloop=1;
	if(op1)
		firstloop=0;
	while(str!=end)
	{
		if(!operand1)
		{
			//If its a constant
			if(scanner->CheckLetterOrNumber(str[0])==1)
				operand1=new Variable(curblock,str);
			//If its a regular variable
			else if(scanner->CheckLetterOrNumber(str[0])==2)
				operand1=GetVariable(str);
			else if(str=="(")
				operand1=DecodeExpression(")",0,"");
			else
				cout << "!Error: A " << str << " is not allowed here!" << endl;
			if(firstloop)
			{
				Variable *tempvar=AddTempVar(operand1->GetType(),
					operand1);
				operand1=tempvar;
				firstloop=0;
			}
		}
		else if(operation=="")
		{
			if(!scanner->CheckLetterOrNumber(str[0]))
			{
				operation=str;
			}
			else
				cout << "!Error: A " << str << " is not allowed here!" << endl;
		}
		else
		{
			//If its a constant
			if(scanner->CheckLetterOrNumber(str[0])==1)
				operand2=new Variable(curblock,str);
			//If its a regular variable
			else if(scanner->CheckLetterOrNumber(str[0])==2)
				operand2=GetVariable(str);
			else if(str=="(")
				operand2=DecodeExpression(")",0,"");
			else
				cout << "!Error: A " << str << " is not allowed here!" << endl;
			if(operation=="+")
			{
				curblock->AddItem(new AddtoCommand(curblock,prog,operand1,
					operand2));
				operand2=0;
				operation="";
			}
			if(operation=="=")
			{
				curblock->AddItem(new AssignCommand(curblock,prog,operand1,
					operand2));
				operand2=0;
				operation="";
			}
		}
		str=scanner->GetNonspaceToken();
	}
	return operand1;
}
*/

int main()
{
	char fname[100];
	gout << "Enter the filename: ";
	cin >> fname;
	CParser p(0,fname);
	gout << "TI-92 assembly:" << endl;
	gout << p.prog->To68kAsm();
	return 0;
}

int CParser::SetFile(CString fname)
{
	if(fname=="")
		return 0;
	scanner=new CScanner(fname);
	if(!scanner->FileOpen())
	{
		delete scanner;
		scanner=0;
		return 0;
	}
	StartParsing();
	return 1;
}

CString CParser::To68kAsm()
{
	return prog->To68kAsm();
}

char CParser::CheckFunctionCall(CString funcname, CString paren)
{
	if(paren!='(')
		return 0;
	CFunction *func=prog->FindItem(funcname);
	if(!func)
		return 0;
	CFunctionCall *call=new CFunctionCall(curblock,prog,func);
	gout << "Its a function call!" << endl;

	int num=0;
	while(scanner->lasttoken!=")")
	{
		CString place="d";
		place+=(char)num+48;
		call->AddParameter(DecodeExpression(",",new CVariable(call,place,1),"=",")"));
		num++;
	}
	CString semicolon=scanner->GetNonspaceToken();
	if(semicolon!=";")
		gout << "Error: ; expected!" << endl;
	curblock->AddItem(call);
	return 1;
}	

char CParser::CheckAsmcode(CString todecode)
{
	if(todecode!="__asm")
		return 0;
	gout << "Its an asm block!" << endl;
	todecode=scanner->GetNonspaceToken();
	CString end;
	if(todecode=="{")
	{
		end="}";
		todecode=scanner->GetNonspaceToken();
	}
	else
		end=";";
	CAsmCommand *asmc=new CAsmCommand(curblock,prog);
	while(todecode!=end)
	{
		asmc->AddAsm(todecode);
		todecode=scanner->GetToken();
	}
	curblock->AddItem(asmc);
	return 1;
}

char CParser::CheckLabel(CString str1, CString str2)
{
	if(str2!=":")
		return 0;
	if(str1=="")
		return 0;
	curblock->AddItem(new CLabelCommand(curblock,prog,str1));
	return 1;
}

char CParser::CheckGoto(CString str1, CString str2)
{
	if(str1!="goto")
		return 0;
	gout << "goto found!" << endl;
	CLabelCommand *label=prog->FindLabel(str2);
	if(!label)
	{
		gout << "Error: " << str2 << " is an undefined label." << endl;
		return 0;
	}
	CString str3=scanner->GetNonspaceToken();
	if(str3!=";")
	{
		gout << "Error: ';' expected!" << endl;
	}
	curblock->AddItem(new CGotoCommand(curblock,prog,label));
	return 1;
}
